###################################################################################################
#                              MIT Licence (C) 2022 Cubicpath@Github                              #
###################################################################################################
"""Initialize values and runs the application."""
import os
import sys
from sys import platform
from typing import Final

import toml
from PySide6.QtCore import *
from toml.decoder import CommentValue

from ._version import __version__
from .client import Client
from .constants import *
from .gui import AppWindow
from .gui import GetterApp
from .tomlfile import *

__all__ = (
    'run',
)


def _make_comment_val(val: TOML_VALUE, comment: str, new_line=False) -> CommentValue:
    return CommentValue(val=val, comment=f'# {comment}', beginline=new_line, _dict=dict)


DEFAULT_SETTINGS = {
    'language': 'en-US',
    'gui': {
        'window': {
            'x_size': _make_comment_val(900, 'Minimum value of 100'),
            'y_size': _make_comment_val(500, 'Minimum value of 100')
        },
        'themes': {
            'selected': 'light',
            'dark': {
                'display_name': 'Breeze Dark',
                'path': RESOURCE_PATH / 'themes/dark'
            },
            'light': {
                'display_name': 'Breeze Light',
                'path': RESOURCE_PATH / 'themes/light'
            }
        },
        'media_output': {
            'aspect_ratio_mode': _make_comment_val(1, '0: Ignore | 1: Keep | 2: Expanding'),
            'transformation_mode': _make_comment_val(0, '0: Fast | 1: Smooth')
        },
        'text_output': {
            'line_wrap_mode': _make_comment_val(1, '0: No Wrap | 1: Widget | 2: Fixed Pixel | 4: Fixed Column')
        }
    }
}
"""Default settings to use in config."""


def _patch_windows_taskbar_icon() -> None:
    """Override Python's default Windows taskbar icon with the custom one set by the app window."""
    if platform == 'win32':
        from ctypes import windll
        windll.shell32.SetCurrentProcessExplicitAppUserModelID(f'cubicpath.hi_getter.app.{__version__}')


def run(*args, **kwargs) -> int:
    """Run the program.

    Args are passed to a QApplication instance.
    Kwargs are handled here.
    """

    settings_path = (CONFIG_PATH / 'settings.toml')

    _patch_windows_taskbar_icon()

    # Create files and paths if they do not exist
    if not CONFIG_PATH.is_dir():
        os.makedirs(CONFIG_PATH)
    if not settings_path.is_file():
        with settings_path.open(mode='w', encoding='utf8') as file:
            toml.dump(DEFAULT_SETTINGS, file, encoder=BetterTomlEncoder())

    APP:        Final[GetterApp] = GetterApp(list(args), TomlFile(settings_path, default=DEFAULT_SETTINGS))
    APP.load_themes()

    CLIENT:     Final[Client] = kwargs.pop('client', Client())
    SIZE:       Final[QSize] = QSize(
        # Size to use, with a minimum of 100x100.
        max(kwargs.pop('x_size', APP.settings['gui/window/x_size']), 100),
        max(kwargs.pop('y_size', APP.settings['gui/window/y_size']), 100)
    )
    WINDOW:     Final[AppWindow] = AppWindow(CLIENT, APP, SIZE)
    WINDOW.show()
    return APP.exec()


if __name__ == '__main__':
    run(sys.argv)



###################################################################################################
#                              MIT Licence (C) 2022 Cubicpath@Github                              #
###################################################################################################
"""Module containing miscellaneous :py:class:`QWidget` Widgets."""
from PySide6.QtCore import *
from PySide6.QtGui import *
from PySide6.QtWidgets import *

from ..constants import *

__all__ = (
    'LicenseViewer',
)


class LicenseViewer(QWidget):
    """Widget that formats and shows the project's license file."""

    def __init__(self) -> None:
        """Create a new LicenseViewer. Gets license from the RESOURCE_PATH/LICENSE file

        Has a fixed size of 750x380.
        """
        super().__init__()
        self.setWindowTitle('License Viewer')
        self.setWindowIcon(QIcon(str(RESOURCE_PATH / 'icons/copyright.ico')))
        self.resize(QSize(750, 380))
        self.setFixedSize(self.size())

        with (RESOURCE_PATH / 'LICENSE').open(mode='r', encoding='utf8') as file:
            license_text = QTextEdit()
            for line in file.readlines():
                license_text.append(line)

        layout = QHBoxLayout()
        self.setLayout(layout)

        layout.addWidget(license_text)

        block_format = QTextBlockFormat()
        block_format.setLineHeight(40, QTextBlockFormat.LineHeightTypes.ProportionalHeight)
        cursor = license_text.textCursor()
        cursor.clearSelection()
        cursor.select(QTextCursor.SelectionType.Document)
        cursor.mergeBlockFormat(block_format)
        license_text.setFont(QFont('consolas', 12))
        license_text.setDisabled(True)



###################################################################################################
#                              MIT Licence (C) 2022 Cubicpath@Github                              #
###################################################################################################
"""Loads data and defines the HTTP Client."""
import json
import os
import random
import time
from pathlib import Path
from typing import Any

from dotenv import load_dotenv
from requests import Response
from requests import Session
from requests.utils import guess_json_utf

from .constants import CONFIG_PATH
from .utils import *

__all__ = (
    'Client',
)

load_dotenv(verbose=True)


class Client:
    """HTTP REST Client that interfaces with HaloWaypoint to get data."""
    HOST: str = 'svc.halowaypoint.com'
    PARENT_PATH: str = '/hi/'

    def __init__(self, **kwargs) -> None:
        """Initializes HaloWaypoint Client

        API Key is first taken from auth kwarg, then SPARTAN_AUTH environment variable, then from the user's config directory.

        :keyword auth: Token to authenticate self to 343 API.
        """
        user_key_path = CONFIG_PATH / 'api_key'
        self._auth: str = kwargs.pop('auth', os.getenv('SPARTAN_AUTH', user_key_path.read_text(encoding='utf8').strip() if user_key_path.is_file() else None))
        self.sub_host: str = 'gamecms-hacs-origin'  # Must be defined before session headers

        self.searched_paths = {}
        self.session: Session = Session()
        self.session.headers = {
            'Accept': ', '.join(('application/json', 'text/plain', '*/*')),
            'Accept-Encoding': ', '.join(('gzip', 'deflate', 'br')),
            'Accept-Language': 'en-US',
            'Connection': 'keep-alive',
            'Cache-Control': 'no-cache',
            'DNT': '1',
            'Host': self.host,
            'Origin': 'https://www.halowaypoint.com',
            'Pragma': 'no-cache',
            'Referer': 'https://www.halowaypoint.com/',
            'Sec-Fetch-Dest': 'empty',
            'Sec-Fetch-Mode': 'cors',
            'Sec-Fetch-Site': 'same-site',
            'Sec-GPC': '1',
            'TE': 'trailers',
            'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:100.0) Gecko/20100101 Firefox/100.0',
            'x-343-authorization-spartan': self._auth
        }
        self.session.cookies.set('343-spartan-token', self._auth)
        data_path = Path.cwd() / 'hi_data'
        if not data_path.is_dir():
            os.mkdir(data_path)

    # def _refresh_auth(self) -> None:
    #     self._AUTH = ...

    def get(self, path: str, **kwargs) -> Response:
        """Get a :py:class:`Response` from HaloWaypoint and update cookies.

        :param path: path to append to the API root
        :param kwargs: Key word arguments to pass to the requests GET Request.
        """
        response: Response = self.session.get(self.api_root + path.strip(), **kwargs)
        self.session.cookies.update(response.cookies)
        if not response.ok:
            if response.status_code == 401:
                ...
                # TODO: Automatically refresh api key
                # print('test')
                # self._refresh_auth()
                # time.sleep(1)
                # response = self.get(path, **kwargs)
        return response

    def get_hi_data(self, path: str, only_dump: bool = False, dump_path: Path = Path.cwd()) -> dict[str, Any] | bytes | int | None:
        """Returns data from a path. Return type depends on the resource.

        :return: dict for JSON objects, bytes for media, int for error codes.
        """
        os_path: Path = dump_path / Path(('hi_data/' + path.replace('/file/', '/')).lower())
        data: dict[str, Any] | bytes | int | None = None

        if not os_path.is_file():
            response: Response = self.get(path)
            if not response.ok:
                return response.status_code
            if 'json' in response.headers.get('content-type'):
                data = response.json()
            else:
                data = response.content

            print(f"DOWNLOADED {path} >>> {response.headers.get('content-type')}")
            dump_data(os_path, data)
            time.sleep(random.randint(100, 200) / 750)

        elif not only_dump:
            print(path)
            data = os_path.read_bytes()
            if os_path.suffix == '.json':
                data = json.loads(data.decode(guess_json_utf(data)))

        return data

    def recursive_search(self, search_path: str) -> None:
        """Recursively get Halo Infinite files linked to the search_path through Mapping keys."""
        if self.searched_paths.get(search_path, 0) >= 2:
            return

        data: dict[str, Any] | bytes | int = self.get_hi_data(search_path)
        if isinstance(data, (bytes, int)):
            return

        for value in unique_values(data):
            if isinstance(value, str):
                if '/' not in value:
                    continue

                end = value.split('.')[-1].lower()
                if end in ('json',):
                    path = 'progression/file/' + value
                    self.searched_paths.update({path: self.searched_paths.get(path, 0) + 1})
                    self.recursive_search(path)
                elif end in ('png', 'jpg', 'jpeg', 'webp', 'gif'):
                    self.get_hi_data('images/file/' + value, True)

    @property
    def auth_key(self) -> str:
        """343 auth token to authenticate self to HaloWaypoint."""
        return self._auth

    @auth_key.setter
    def auth_key(self, value: str) -> None:
        self._auth = value
        self.session.headers.update({'x-343-authorization-spartan': self._auth})
        self.session.cookies.set('343-spartan-token', self._auth)

    @property
    def api_root(self) -> str:
        """Root of sent API requests."""
        return f'https://{self.host}{self.PARENT_PATH}'

    @property
    def host(self) -> str:
        """Host to send to."""
        return f'{self.sub_host}.{self.HOST}'

    @host.setter
    def host(self, value: str) -> None:
        self.sub_host = value



###################################################################################################
#                              MIT Licence (C) 2022 Cubicpath@Github                              #
###################################################################################################
"""Module containing :py:class:`QMenu` Context Menus."""
import sys
import webbrowser
from importlib import metadata
from platform import platform
from shutil import rmtree

import requests
from PySide6.QtGui import *
from PySide6.QtWidgets import *

from .._version import __version_info__ as ver
from ..constants import *
from .widgets import LicenseViewer

__all__ = (
    'HelpContextMenu',
    'FileContextMenu',
)


# noinspection PyArgumentList
class HelpContextMenu(QMenu):
    """Context menu that shows actions to help the user."""

    _about: tuple[str, str] | None = None
    _github_icon: bytes | None = None

    def __init__(self, parent) -> None:
        """Create a new :py:class:`HelpContextMenu`."""
        super().__init__(parent)

        self.license_window = LicenseViewer()

        if self._github_icon is None:
            self.__class__._github_icon = requests.get('https://github.githubassets.com/favicons/favicon.png').content

        ico_pixmap = QPixmap()
        ico_pixmap.loadFromData(self._github_icon)

        github_view = QAction(QIcon(ico_pixmap), 'View on GitHub', self, triggered=self.open_github)
        create_issue = QAction(QIcon(ico_pixmap), 'Create an issue', self, triggered=self.open_issues)
        license_view = QAction(QIcon(str(RESOURCE_PATH / 'icons/copyright.ico')), 'License', self, triggered=self.open_license)
        about_view = QAction(QIcon(str(RESOURCE_PATH / 'icons/about.ico')), 'About', self, triggered=self.open_about)
        self.addSection('Github')
        self.addAction(github_view)
        self.addAction(create_issue)
        self.addSection('Information')
        self.addAction(license_view)
        self.addAction(about_view)

    @staticmethod
    def open_github() -> None:
        """Open the project's GitHub repository in the user's default browser."""
        webbrowser.open('https://github.com/Cubicpath/HaloInfiniteGetter/', new=2, autoraise=True)

    @staticmethod
    def open_issues() -> None:
        """Creates a new GitHub issue using the user's default browser."""
        webbrowser.open('https://github.com/Cubicpath/HaloInfiniteGetter/issues/new', new=2, autoraise=True)

    def open_license(self) -> None:
        """Show the :py:class:`LicenseViewer` window."""
        self.license_window.show()

    # pylint: disable=not-an-iterable
    def open_about(self) -> None:
        """Open the application's about section."""
        self.setWindowIcon(QIcon(str(RESOURCE_PATH / 'icons/about.ico')))
        QMessageBox.information(self, *self.about_message())

    def about_message(self) -> tuple[str, str]:
        """Generate the 'About' information regarding the application's environment.
        Since this takes a while the first time, the result is cached as a class value.

        :return: Tuple containing the title and the message
        """
        if self._about is None:
            self.__class__._about = ('About', f'''
HaloInfiniteGetter by Cubicpath. A simple way to get live Halo data straight from Halo Waypoint.

Version Info: major:{ver[0]}, minor:{ver[1]}, micro:{ver[2]}, releaselevel:{ver[3]}, serial:{ver[4]}

Running on:
\t{platform()}
\t{sys.implementation.name} {sys.version.split('[', maxsplit=1)[0]}

Using requests: {metadata.version("requests")}
Using PySide6: {metadata.version("PySide6")}
Using python-dotenv: {metadata.version("python-dotenv")}

MIT Licence (C) 2022 Cubicpath@Github
''')
        return self._about


# noinspection PyArgumentList
class FileContextMenu(QMenu):
    """Context menu that shows actions to manipulate files."""
    def __init__(self, parent) -> None:
        """Create a new :py:class:`FileContextMenu`."""
        super().__init__(parent)

        open_in:     QAction = QAction(QIcon(str(RESOURCE_PATH / 'icons/folder.ico')), 'Open In Explorer', self, triggered=self.open_explorer)
        flush_cache: QAction = QAction(QIcon(str(RESOURCE_PATH / 'icons/folder.ico')), 'Flush Cache', self, triggered=self.flush_cache)
        import_from: QAction = QAction(QIcon(str(RESOURCE_PATH / 'icons/import.ico')), 'Import Data from...', self, triggered=self.import_data)
        export_to:   QAction = QAction(QIcon(str(RESOURCE_PATH / 'icons/export.ico')), 'Export Data', self, triggered=self.export_data)
        self.addAction(open_in)
        self.addAction(flush_cache)
        self.addAction(import_from)
        self.addAction(export_to)

        flush_cache.setDisabled(tuple(CACHE_PATH.iterdir()) == ())  # Set disabled if CACHE_PATH directory is empty
        # TODO: Add functionality and enable
        import_from.setDisabled(True)
        export_to.setDisabled(True)

    @staticmethod
    def open_explorer() -> None:
        """Open the user's file explorer in the current application's export directory."""
        webbrowser.open(f'file:///{CACHE_PATH}')

    @staticmethod
    def flush_cache() -> None:
        """Remove all data from cache."""
        rmtree(CACHE_PATH)
        CACHE_PATH.mkdir()

    def import_data(self) -> None:
        """Import data from..."""
        ...

    def export_data(self) -> None:
        """Export data to..."""
        ...



###################################################################################################
#                              MIT Licence (C) 2022 Cubicpath@Github                              #
###################################################################################################
"""Neutral namespace for constants."""
import re
from pathlib import Path
from typing import Final

__all__ = (
    'CACHE_PATH',
    'CONFIG_PATH',
    'PATH_PATTERN',
    'RESOURCE_PATH',
)

PATH_PATTERN:  Final[re.Pattern] = re.compile(r'\"([\w\-_]+/)+[\w\-_]*\.\w+\"')
"""Regex pattern for finding a resource path. Finds quoted substrings with at least one folder name and file name (with a file extension)."""

CACHE_PATH:    Final[Path] = Path.cwd() / 'hi_data'
"""Directory containing cached API results."""

CONFIG_PATH:   Final[Path] = Path.home() / '.config/hi_getter'
"""Directory containing user configuration data."""

RESOURCE_PATH: Final[Path] = Path(__file__).parent / 'resources'
"""Directory containing application resources."""



###################################################################################################
#                              MIT Licence (C) 2022 Cubicpath@Github                              #
###################################################################################################
"""Module containing GUI elements meant to be used as windows."""
import json
import webbrowser
from collections.abc import Callable
from pathlib import Path

from PySide6.QtCore import *
from PySide6.QtGui import *
from PySide6.QtWidgets import *

from .._version import __version__
from ..client import Client
from ..constants import *
from ..tomlfile import *
from ..utils import HTTP_CODE_MAP
from .app import *
from .menus import *

__all__ = (
    'AppWindow',
    'SettingsWindow',
)


# noinspection PyArgumentList
class SettingsWindow(QWidget):
    """Window that provides user interaction with the application's settings."""

    def __init__(self, parent: 'AppWindow', size: QSize) -> None:
        """Create a new settings window. Should only have one instance."""
        super().__init__()
        self.client = parent.client
        self.clipboard: QClipboard | None = parent.clipboard
        self.getter_window = parent
        self.settings: TomlFile = parent.APP.settings
        self.setWindowTitle('Settings')
        self.setWindowIcon(QIcon(str(RESOURCE_PATH / 'icons/settings.ico')))
        self.resize(size)
        self.setFixedWidth(self.width())
        self.settings.hook_event('$fail:import', lambda: QMessageBox.warning(
            self, 'Import Error',
            f'Could not import settings from "{self.settings.path}". Make sure both the file and '
            f'the parent folder exist, and that the file contains valid TOML.'
        ))

        self.save_button:             QPushButton
        self.theme_dropdown:          QComboBox
        self.aspect_ratio_dropdown:   QComboBox
        self.transformation_dropdown: QComboBox
        self.line_wrap_dropdown:      QComboBox
        self.key_set_button:          QPushButton
        self.key_field:               QLineEdit
        self._init_ui()

    def _init_ui(self) -> None:
        self.save_button:             QPushButton = QPushButton('Save', clicked=self.save_settings)
        reload_button:                QPushButton = QPushButton('Reload', clicked=self.reload_settings)
        import_button:                QPushButton = QPushButton('Import Settings', clicked=self.import_settings)
        export_button:                QPushButton = QPushButton('Export Settings', clicked=self.export_settings)
        theme_label:                  QLabel = QLabel('Current Theme: ')
        aspect_ratio_label:           QLabel = QLabel('Aspect Ratio: ')
        transformation_label:         QLabel = QLabel('Image Transform: ')
        line_wrap_label:              QLabel = QLabel('Line Wrap: ')
        self.theme_dropdown:          QComboBox = QComboBox(self)
        self.aspect_ratio_dropdown:   QComboBox = QComboBox(self)
        self.transformation_dropdown: QComboBox = QComboBox(self)
        self.line_wrap_dropdown:      QComboBox = QComboBox(self)
        open_editor_button:           QPushButton = QPushButton('Open Settings in Editor', clicked=self.open_editor)
        key_show_button:              QPushButton = QPushButton('Edit Auth Key', clicked=self.show_key)
        key_copy_button:              QPushButton = QPushButton('Copy to Clipboard', clicked=self.copy_key)
        self.key_set_button:          QPushButton = QPushButton('Set', clicked=self.set_key)
        self.key_field:               QLineEdit = QLineEdit(returnPressed=self.key_set_button.click)

        self.theme_dropdown.activated.connect(self.set_theme)
        self.transformation_dropdown.activated.connect(self.set_transformation_method)
        self.line_wrap_dropdown.activated.connect(self.set_line_wrap_method)
        self.aspect_ratio_dropdown.activated.connect(self.set_aspect_ratio_method)

        # Define layouts
        layout = QGridLayout()  # Main layout
        top = QHBoxLayout()
        middle = QVBoxLayout()
        theme_layout = QHBoxLayout()
        output_layout = QGridLayout()
        bottom = QVBoxLayout()
        key_layout = QHBoxLayout()

        # Assign positions of layouts
        self.setLayout(layout)
        layout.addLayout(top, 0, 0, Qt.AlignTop)
        layout.addLayout(middle, 10, 0, Qt.AlignTop)
        layout.addLayout(bottom, 20, 0, Qt.AlignBottom)

        # Add top widgets
        top.addWidget(self.save_button)
        top.addWidget(reload_button)
        top.addWidget(import_button)
        top.addWidget(export_button)

        # Add middle widgets
        middle.addWidget(open_editor_button)
        middle.addLayout(theme_layout)
        middle.addLayout(output_layout)
        theme_layout.addWidget(theme_label)
        theme_layout.addWidget(self.theme_dropdown)
        output_layout.addWidget(aspect_ratio_label, 0, 0)
        output_layout.addWidget(self.aspect_ratio_dropdown, 0, 10)
        output_layout.addWidget(transformation_label, 0, 20)
        output_layout.addWidget(self.transformation_dropdown, 0, 30)
        output_layout.addWidget(line_wrap_label, 10, 0)
        output_layout.addWidget(self.line_wrap_dropdown, 10, 10)

        # Add bottom widgets
        bottom.addWidget(key_show_button)
        bottom.addLayout(key_layout)
        key_layout.addWidget(key_copy_button)
        key_layout.addWidget(self.key_field)
        key_layout.addWidget(self.key_set_button)

        # Modify properties of widgets
        theme_label.setMaximumWidth(85)

        self.theme_dropdown.addItems([theme.display_name for theme in self.getter_window.APP.sorted_themes])
        self.theme_dropdown.setCurrentIndex(self.getter_window.APP.theme_index_map[self.settings['gui/themes/selected']])

        aspect_ratio_label.setMaximumWidth(70)
        self.aspect_ratio_dropdown.addItems(('Ignore', 'Keep', 'Expanding'))
        self.aspect_ratio_dropdown.setCurrentIndex(self.settings['gui/media_output/aspect_ratio_mode'])
        transformation_label.setMaximumWidth(90)
        self.transformation_dropdown.addItems(('Fast', 'Smooth'))
        self.transformation_dropdown.setCurrentIndex(self.settings['gui/media_output/transformation_mode'])
        self.line_wrap_dropdown.addItems(('No Wrap', 'Widget', 'Fixed Pixel', 'Fixed Column'))
        self.line_wrap_dropdown.setCurrentIndex(self.settings['gui/text_output/line_wrap_mode'])

        self.save_button.setMaximumWidth(50)
        self.save_button.setDisabled(True)
        reload_button.setMaximumWidth(60)

        self.key_field.setAlignment(Qt.AlignCenter)
        self.key_field.setMinimumWidth(220)
        self.key_field.setFont(QFont('segoe ui', 8))
        self.key_set_button.setMinimumWidth(40)

    def refresh_dropdowns(self) -> None:
        """Refresh all dropdown widgets with the current settings assigned to them."""
        self.aspect_ratio_dropdown.setCurrentIndex(self.settings['gui/media_output/aspect_ratio_mode'])
        self.transformation_dropdown.setCurrentIndex(self.settings['gui/media_output/transformation_mode'])
        self.line_wrap_dropdown.setCurrentIndex(self.settings['gui/text_output/line_wrap_mode'])
        self.theme_dropdown.setCurrentIndex(self.getter_window.APP.theme_index_map[self.settings['gui/themes/selected']])

    def save_settings(self) -> None:
        """Save current settings to the user's settings file."""
        self.save_button.setDisabled(True)
        self.settings.save()

    def reload_settings(self) -> None:
        """Reload current settings from the user's settings file."""
        self.save_button.setDisabled(True)
        if self.settings.reload():
            self.refresh_dropdowns()

    def import_settings(self) -> None:
        """Import settings from a chosen TOML file."""
        self.save_button.setDisabled(True)
        file_path = Path(QFileDialog.getOpenFileName(self, 'Import Settings', str(CONFIG_PATH), 'TOML Files (*.toml);;All files (*.*)')[0])
        if file_path.is_file():
            if self.settings.import_from(file_path):
                self.refresh_dropdowns()

    def export_settings(self) -> None:
        """Export current settings to a chosen file location."""
        file_path = Path(QFileDialog.getSaveFileName(self, 'Export Settings', str(CONFIG_PATH), 'TOML Files (*.toml);;All files (*.*)')[0])
        if str(file_path) != '.':
            self.settings.export_to(file_path)

    def open_editor(self) -> None:
        """Open current settings file in the user's default text editor."""
        webbrowser.open(str(self.settings.path))

    def set_aspect_ratio_method(self) -> None:
        """Set the media output's aspect ratio method to the chosen method."""
        self.save_button.setDisabled(False)
        self.settings['gui/media_output/aspect_ratio_mode'] = self.aspect_ratio_dropdown.currentIndex()
        self.getter_window.resize_image()

    def set_transformation_method(self) -> None:
        """Set the media output's image transformation method to the chosen method."""
        self.save_button.setDisabled(False)
        self.settings['gui/media_output/transformation_mode'] = self.transformation_dropdown.currentIndex()
        self.getter_window.resize_image()

    def set_line_wrap_method(self) -> None:
        """Set the text output's line wrap method to the chosen method."""
        self.save_button.setDisabled(False)
        self.settings['gui/text_output/line_wrap_mode'] = self.line_wrap_dropdown.currentIndex()
        self.getter_window.text_output.setLineWrapMode(QTextEdit.LineWrapMode(self.settings['gui/text_output/line_wrap_mode']))

    def set_theme(self) -> None:
        """Set selected theme to the chosen theme."""
        self.save_button.setDisabled(False)
        self.settings['gui/themes/selected'] = self.getter_window.APP.sorted_themes[self.theme_dropdown.currentIndex()].id

    def hide_key(self) -> None:
        """Hide API key."""
        self.key_set_button.setDisabled(True)
        self.key_field.setDisabled(True)
        self.key_field.setText(self.hidden_key())
        self.key_field.setAlignment(Qt.AlignCenter)

    def show_key(self) -> None:
        """Toggle hiding and showing the API key."""
        if not self.key_field.isEnabled():
            self.key_field.setAlignment(Qt.AlignLeft)
            self.key_field.setText(self.client.auth_key)
            self.key_field.setDisabled(False)
            self.key_field.setFocus()
            self.key_set_button.setDisabled(False)
        else:
            self.hide_key()

    def copy_key(self) -> None:
        """Copy the current key value to the system clipboard."""
        if self.clipboard is not None:
            self.clipboard.setText(self.client.auth_key)

    def set_key(self) -> None:
        """Set the client's auth_key to the current text in the key field."""
        self.client.auth_key = self.key_field.text().strip().removeprefix('x-343-authorization-spartan: ')
        self.show_key()

    def hidden_key(self) -> str:
        """:return: The first 5 and last 4 characters of the API key, seperated by periods."""
        key = self.client.auth_key
        if key is not None and len(key) > 10:
            return f'{key[:5]}{"." * 50}{key[-4:]}'
        return 'None'

    # # # # # Events

    def showEvent(self, event: QShowEvent) -> None:
        """Auto hides the key upon un-minimizing."""
        super().showEvent(event)
        self.key_set_button.setDisabled(True)
        self.key_field.setAlignment(Qt.AlignCenter)
        self.key_field.setDisabled(True)
        self.key_field.setText(self.hidden_key())


# noinspection PyArgumentList
class AppWindow(QMainWindow):
    """Main window for the HaloInfiniteGetter application."""

    shown_key_warning: bool = False

    def __init__(self, client: Client, app: GetterApp, size: QSize) -> None:
        """Create the window for the application."""
        super().__init__()
        self.APP:                   GetterApp = app
        self.client:                Client = client
        self.clipboard:             QClipboard = app.clipboard()
        self.current_image:         QPixmap | None = None
        self._clicked_input_field:  bool = False
        self.detached:              dict[str, QMainWindow | None] = {'media': None, 'text': None}
        # self.themes: dict[str, str] = themes
        self.setWindowTitle(f'HaloInfiniteGetter v{__version__}')
        self.setWindowIcon(QIcon(str(RESOURCE_PATH / 'icons/hi.ico')))
        self.resize(size)

        self.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)

        self.input_field:   QLineEdit
        self.media_output:  QGraphicsView
        self.text_output:   QTextBrowser
        self.clear_picture: QPushButton
        self.copy_picture:  QPushButton
        self.clear_text:    QPushButton
        self.copy_text:     QPushButton

        self._init_toolbar()
        self._init_ui()

        self.settings_window = SettingsWindow(self, QSize(420, 600))

    def _init_toolbar(self) -> None:
        """Initialize toolbar widgets."""
        self.toolbar = QToolBar('Toolbar', self)
        file = QAction('File', self, triggered=self.file_context_handler)
        settings = QAction('Settings', self, triggered=self.open_settings_window)
        help_ = QAction('Help', self, triggered=self.help_context_handler)

        self.addToolBar(self.toolbar)
        for action in (file, settings, help_):
            self.toolbar.addSeparator()
            self.toolbar.addAction(action)

        file.setMenuRole(QAction.MenuRole.ApplicationSpecificRole)
        settings.setMenuRole(QAction.MenuRole.PreferencesRole)
        help_.setMenuRole(QAction.MenuRole.AboutRole)

    def _init_ui(self) -> None:
        """Initialize the UI, including Layouts and widgets."""
        self.media_frame = QFrame()
        self.image_size_label = QLabel('Image Output:')
        self.image_detach_button = QPushButton('Detach', clicked=self.toggle_media_detach)
        self.media_output = QGraphicsView()

        self.text_frame = QFrame()
        self.text_size_label = QLabel('Text Output:')
        self.text_detach_button = QPushButton('Detach', clicked=self.toggle_text_detach)
        self.text_output = QTextBrowser()
        self.text_output.anchorClicked.connect(lambda e: self.navigate_to(e.toDisplayString()))

        self.clear_picture = QPushButton('Clear', clicked=self.clear_current_pixmap)
        self.copy_picture = QPushButton('Copy Picture', clicked=self.copy_current_pixmap)
        self.clear_text = QPushButton('Clear', clicked=self.clear_current_text)
        self.copy_text = QPushButton('Copy Text', clicked=self.copy_current_text)

        subdomain_field = QLineEdit(self.client.sub_host)
        root_folder_field = QLineEdit(self.client.PARENT_PATH)
        get_button = QPushButton('GET', clicked=self.get_resource)
        scan_button = QPushButton('SCAN', clicked=self.scan_resource)

        self.input_field = QLineEdit('Progression/file/Calendars/Seasons/SeasonCalendar.json', returnPressed=get_button.click)

        main_widget = QWidget()
        layout = QGridLayout()
        top = QHBoxLayout()
        self.outputs = QHBoxLayout()
        media_layout = QVBoxLayout()
        media_top = QHBoxLayout()
        media_bottom = QHBoxLayout()
        text_layout = QVBoxLayout()
        text_top = QHBoxLayout()
        text_bottom = QHBoxLayout()
        bottom = QGridLayout()

        self.setCentralWidget(main_widget)
        main_widget.setLayout(layout)
        layout.addLayout(top, 0, 0, Qt.AlignTop)
        layout.addLayout(self.outputs, 10, 0, Qt.AlignHCenter)
        layout.addLayout(bottom, 20, 0, Qt.AlignBottom)

        top.addWidget(subdomain_field)
        top.addWidget(root_folder_field)
        top.addWidget(self.input_field)
        top.addWidget(get_button)
        top.addWidget(scan_button)
        top.setSpacing(2)

        self.outputs.addWidget(self.media_frame)
        self.outputs.addWidget(self.text_frame)

        # noinspection Duplicates
        self.media_frame.setLayout(media_layout)
        media_layout.addLayout(media_top)
        media_layout.addWidget(self.media_output)
        media_layout.addLayout(media_bottom)
        media_top.addWidget(self.image_size_label, Qt.AlignLeft)
        media_top.addWidget(self.image_detach_button, Qt.AlignRight)
        media_bottom.addWidget(self.clear_picture, Qt.AlignLeft)
        media_bottom.addWidget(self.copy_picture, Qt.AlignLeft)

        # noinspection Duplicates
        self.text_frame.setLayout(text_layout)
        text_layout.addLayout(text_top)
        text_layout.addWidget(self.text_output)
        text_layout.addLayout(text_bottom)
        text_top.addWidget(self.text_size_label, Qt.AlignLeft)
        text_top.addWidget(self.text_detach_button, Qt.AlignRight)
        text_bottom.addWidget(self.clear_text, Qt.AlignLeft)
        text_bottom.addWidget(self.copy_text, Qt.AlignLeft)
        text_bottom.setSpacing(5)

        subdomain_field.setFixedWidth(125)
        subdomain_field.setDisabled(True)
        root_folder_field.setFixedWidth(28)
        root_folder_field.setDisabled(True)
        self.input_field.mousePressEvent = self._destroy_text_on_first_click(self.input_field.mousePressEvent)
        # subdomain_field.returnPressed.connect(lambda *_: self.client.__class__.host.fset(self.client, subdomain_field.text()))
        get_button.setMaximumWidth(40)
        scan_button.setMaximumWidth(55)
        self.image_size_label.setMinimumWidth(50)
        self.image_detach_button.setMaximumWidth(80)
        self.media_output.setScene(QGraphicsScene())
        self.media_output.setMinimumHeight(28)
        self.media_output.setAutoFillBackground(False)
        self.media_output.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.media_output.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.media_output.setDisabled(True)
        self.text_size_label.setMinimumWidth(50)
        self.text_detach_button.setMaximumWidth(80)
        self.text_output.setMinimumHeight(28)
        self.text_output.setLineWrapMode(QTextEdit.LineWrapMode(self.APP.settings['gui/text_output/line_wrap_mode']))
        self.text_output.setOpenLinks(False)
        self.text_output.setDisabled(True)
        self.clear_picture.setMaximumWidth(80)
        self.clear_picture.setMinimumWidth(40)
        self.clear_picture.setDisabled(True)
        self.copy_picture.setMaximumWidth(160)
        self.copy_picture.setMinimumWidth(80)
        self.copy_picture.setDisabled(True)
        self.clear_text.setMaximumWidth(80)
        self.clear_text.setMinimumWidth(40)
        self.clear_text.setDisabled(True)
        self.copy_text.setMaximumWidth(160)
        self.copy_text.setMinimumWidth(80)
        self.copy_text.setDisabled(True)

    def _destroy_text_on_first_click(self, f: Callable) -> Callable:
        """Destroys the sample text in the input field if clicked before it is used."""
        def wrapper(*args, **kwargs):
            if self._clicked_input_field:
                f(*args, **kwargs)
            else:
                self._clicked_input_field = True
                self.input_field.setText('')
        return wrapper

    def _setup_detached_window(self, frame: QFrame, handler: Callable, title: str = None) -> QMainWindow:
        """Set up a detached window, with the layout represented as a :py:class:`QFrame`.

        :param frame: Set the window's central widget as this QFrame.
        :param handler: Callable to execute when closed, to reattach the frame to the parent window.
        :param title: The window title.
        """
        window = QMainWindow(self)
        window.setWindowTitle(title if title is not None else self.windowTitle())
        window.setCentralWidget(frame)
        window.closeEvent = lambda *_: handler() if self.detached['media'] is not None else None
        window.setMinimumHeight(200)
        window.setMinimumWidth(300)
        return window

    def file_context_handler(self) -> None:
        """Create a new :py:class:`FileContextMenu` and show it at the cursor's position."""
        menu = FileContextMenu(self)
        menu.move(self.cursor().pos())
        menu.show()

    def help_context_handler(self) -> None:
        """Create a new :py:class:`HelpContextMenu` and show it at the cursor's position."""
        menu = HelpContextMenu(self)
        menu.move(self.cursor().pos())
        menu.show()

    def open_settings_window(self) -> None:
        """Show the :py:class:`SettingsWindow` and bring it the front."""
        self.settings_window.show()
        self.settings_window.activateWindow()
        self.settings_window.raise_()

    def navigate_to(self, path: str) -> None:
        """Set input field text to path and get resource."""
        self.input_field.setText(path)
        self.get_resource()

    def toggle_media_detach(self) -> None:
        """Handler for detaching and reattaching the media output."""
        if self.detached['media'] is None:
            self.detached['media'] = window = self._setup_detached_window(self.media_frame, self.toggle_media_detach, 'Detached Image Output')
            self.image_detach_button.setText('Reattach')
            window.resizeEvent = lambda *_: self.resize_image()
            window.show()
        else:
            window = self.detached['media']
            self.detached['media'] = None
            window.close()

            self.outputs.insertWidget(0, self.media_frame)
            self.image_detach_button.setText('Detach')

    def toggle_text_detach(self) -> None:
        """Handler for detaching and reattaching the text output."""
        if self.detached['text'] is None:
            self.detached['text'] = window = self._setup_detached_window(self.text_frame, self.toggle_text_detach, 'Detached Text Output')
            self.text_detach_button.setText('Reattach')
            window.show()
        else:
            window = self.detached['text']
            self.detached['text'] = None
            window.close()

            self.outputs.insertWidget(-1, self.text_frame)
            self.text_detach_button.setText('Detach')

    def clear_current_pixmap(self) -> None:
        """Clear the current image from the media output."""
        self.image_size_label.setText('Image Output: ')
        self.clear_picture.setDisabled(True)
        self.copy_picture.setDisabled(True)
        self.media_output.scene().clear()
        self.current_image = None

    def copy_current_pixmap(self) -> None:
        """Copy the current image to the system clipboard."""
        if self.clipboard is not None:
            self.clipboard.setPixmap(self.current_image)

    def clear_current_text(self) -> None:
        """Clear the current text from the text output."""
        self.text_size_label.setText('Text Output: ')
        self.clear_text.setDisabled(True)
        self.copy_text.setDisabled(True)
        self.text_output.setDisabled(True)
        self.text_output.clear()

    def copy_current_text(self) -> None:
        """Copy the current output text to the system clipboard."""
        if self.clipboard is not None:
            self.clipboard.setText(self.text_output.toPlainText())

    def get_resource(self) -> None:
        """Get a single resource from the resource path."""
        self.use_input(op_code=10)

    def scan_resource(self) -> None:
        """Recursively search through the resource path's JSON data for more links to scan, if any."""
        self.use_input(op_code=20)

    def use_input(self, op_code: int = 0) -> None:
        """Use the current input field's text to search through the Client for data.

        Automatically handles media and text data.

        :param op_code: What operation to do. Get: 10 | Scan: 20
        """
        self._clicked_input_field = True
        user_input = self.input_field.text()
        if '/file/' not in user_input:
            if user_input.endswith(('png', 'jpg', 'jpeg', 'webp', 'gif')):
                user_input = f'images/file/{user_input}'
            else:
                user_input = f'progression/file/{user_input}'

        if user_input:
            if op_code == 10:
                data = self.client.get_hi_data(user_input)
                if isinstance(data, bytes):
                    self.clear_picture.setDisabled(False)
                    self.copy_picture.setDisabled(False)
                    self.current_image = QPixmap()
                    self.current_image.loadFromData(data)
                    size = self.current_image.size()

                    self.image_size_label.setText(f'Image Output: {size.width()}x{size.height()} ({round(len(data) / 1024, 4)} KiB)')
                    self.resize_image()
                else:
                    self.clear_text.setDisabled(False)
                    self.copy_text.setDisabled(False)
                    self.text_output.setDisabled(False)

                    if isinstance(data, dict):
                        data = json.dumps(data, indent=2)
                    elif isinstance(data, int):
                        data = f'{data}: {HTTP_CODE_MAP[data][0]}\n{HTTP_CODE_MAP[data][1]}'

                    output = data
                    replaced = set()

                    for match in PATH_PATTERN.finditer(data):
                        match = match[0].replace('"', '')
                        if match not in replaced:
                            output = output.replace(match, f'<a href="{match}" style="color: #2A5DB0">{match}</a>')
                            replaced.add(match)

                    self.text_output.setHtml(f'<body style="white-space: pre-wrap">{output}</body>')
                    self.text_size_label.setText(f'Text Output: '
                                                 f'{len(data.splitlines())} lines; '
                                                 f'{len(data)} characters '
                                                 f'({round(len(data.encode("utf8")) / 1024, 4)} KiB)')
            elif op_code == 20:
                self.client.recursive_search(user_input)
                self.use_input(op_code=10)

    def resize_image(self) -> None:
        """Refresh the media output with a resized version of the current image."""
        if self.current_image is not None:
            new = self.current_image.copy()
            self.media_output.scene().clear()  # Clear buffer, otherwise causes memory leak
            if self.current_image.size() != self.media_output.viewport().size():
                # Create a new image from the source image, scaled to fit the window.
                new = new.scaled(
                    self.media_output.viewport().size(),
                    Qt.AspectRatioMode(self.APP.settings['gui/media_output/aspect_ratio_mode']),
                    Qt.TransformationMode(self.APP.settings['gui/media_output/transformation_mode'])
                )
            self.media_output.scene().addPixmap(new)

    # # # # # Events
    def show(self) -> None:
        """After window is displayed, show warnings if not already warned."""
        super().show()
        if not self.shown_key_warning and self.client.auth_key is None:
            QMessageBox.warning(
                self,
                'Empty API Token', '''
The spartan authorization token is not set, please set SPARTAN_AUTH environment variable and restart, or set the auth key value in Settings.

You will be unable to acquire new data until a new token is provided.''')
            self.__class__.shown_key_warning = True

    def resizeEvent(self, event: QResizeEvent) -> None:
        """Resize image on resize of window."""
        super().resizeEvent(event)
        self.resize_image()

    def closeEvent(self, event: QCloseEvent) -> None:
        """Closes all detached/children windows and quit application."""
        super().closeEvent(event)
        self.APP.quit()



###################################################################################################
#                              MIT Licence (C) 2022 Cubicpath@Github                              #
###################################################################################################
"""Module for the main application classes."""
from pathlib import Path

from PySide6.QtCore import *
from PySide6.QtWidgets import *
from toml.decoder import CommentValue

from ..tomlfile import *

__all__ = (
    'GetterApp',
    'Theme',
)


class Theme:
    """Object containing data about a Theme."""

    def __init__(self, id: str, style: str = '', display_name: str | None = None) -> None:
        """Create a new Theme using a unique ID, the style data (.qss file content), and an optional display name.

        :param id: Uniquely identifying string, must be a valid TOML Table name.
        :param style: String read from a .qss file to use as a stylesheet.
        :param display_name: Optional display name to represent the theme, as opposed to the id.
        """
        self.id:           str = id
        self.style:        str = style
        self.display_name: str = display_name if display_name is not None else self.id


class GetterApp(QApplication):
    """The main HaloInfiniteGetter PySide application that runs in the background and manages the process."""
    _legacy_style: str = None

    def __init__(self, argv: list[str], settings: TomlFile) -> None:
        """Create a new app with the given arguments and settings."""
        super().__init__(argv)
        self.settings:        TomlFile = settings
        self.themes:          dict[str, Theme] = {}  # PyCharm detects this as a dict[str, str], despite explicit type hints.
        self.theme_index_map: dict[str, int] = {}

        self.settings.hook_event('$reload', self.load_themes)
        self.settings.hook_event('$set:gui/themes/selected', self.update_stylesheet)

    def update_stylesheet(self) -> None:
        """Set the application stylesheet to the one currently selected in settings."""
        self.setStyleSheet(self.themes[self.settings['gui/themes/selected']].style)

    def load_themes(self) -> None:
        """Load all theme locations from settings and store them in self.themes.

        Also set current theme from settings.
        """
        if self._legacy_style is None:
            self.__class__._legacy_style = self.styleSheet()
        self.themes['legacy'] = Theme('legacy', self._legacy_style, 'Legacy (Default Qt)')

        for id_, theme in self.settings['gui/themes'].items():
            if not isinstance(theme, dict):
                continue

            theme: dict = theme.copy()
            path = theme.pop('path')
            if isinstance(path, CommentValue):
                path = Path(path.val)
            if path.is_dir():
                QDir.addSearchPath(path.name, str(path))

                file = QFile(f'{path.name}:stylesheet.qss')
                file.open(QFile.OpenModeFlag.ReadOnly | QFile.OpenModeFlag.Text)
                theme['id'] = id_
                theme['style'] = QTextStream(file).readAll()
                self.themes[id_] = Theme(**theme)
                file.close()

        # noinspection PyUnresolvedReferences
        self.theme_index_map = {theme_id: i for i, theme_id in enumerate(theme.id for theme in self.sorted_themes)}
        self.update_stylesheet()

    @property
    def sorted_themes(self) -> list[Theme]:
        """List of themes sorted by their display name."""
        # noinspection PyTypeChecker
        return sorted(self.themes.values(), key=lambda theme: theme.display_name)



###################################################################################################
#                              MIT Licence (C) 2022 Cubicpath@Github                              #
###################################################################################################
"""Module used for TOML configurations."""
import warnings
from collections.abc import Callable
from pathlib import Path
from pathlib import PurePath
from types import UnionType
from typing import Any

import toml
from toml.decoder import CommentValue
from toml.decoder import TomlDecodeError
from toml.decoder import TomlPreserveCommentDecoder
from toml.encoder import _dump_str
from toml.encoder import TomlEncoder
# noinspection PyProtectedMember

__all__ = (
    'BetterTomlDecoder',
    'BetterTomlEncoder',
    'TOML_VALUE',
    'TomlFile',
)

TOML_VALUE: UnionType = dict | list | float | int | str | bool | PurePath
"""Represents a possible TOML value, with :py:class:`dict` being a Table, and :py:class:`list` being an Array."""

SPECIAL_PATH_PREFIX = '$PATH$|'

# TODO: Move event_subscribers to an EventBus class


class BetterTomlDecoder(TomlPreserveCommentDecoder):
    """Inherits the effects of :py:class:`toml.TomlPreserveCommentEncoder`.

    With native support for pathlib :py:class:`Path` values; not abandoning the TOML specification.
    """
    def load_value(self, v: str, strictly_valid=True):
        """If the value is a string and starts with the SPECIAL_PATH_PREFIX, load the value enclosed in quotes as a :py:class:`Path`."""
        if v[1:].startswith(SPECIAL_PATH_PREFIX):
            v = Path(v[1:].removeprefix(SPECIAL_PATH_PREFIX)[:-1])
            return v, strictly_valid
        return super().load_value(v=v, strictly_valid=strictly_valid)


class BetterTomlEncoder(TomlEncoder):
    """Combines both the effects of :py:class:`toml.TomlPreserveCommentEncoder` and of :py:class:`toml.TomlPathlibEncoder`.

    Has native support for pathlib :py:class:`PurePath`; not abandoning the TOML specification.
    """
    def __init__(self, _dict=dict, preserve=False):
        super().__init__(_dict, preserve)
        self.dump_funcs[CommentValue] = lambda comment_val: comment_val.dump(self.dump_value)
        self.dump_funcs[PurePath] = self._dump_pathlib_path

    @staticmethod
    def _dump_pathlib_path(v: PurePath):
        """Translate :py:class:`PurePath` to string and dump."""
        return _dump_str(str(v))

    def dump_value(self, v: TOML_VALUE):
        """Support :py:class:`Path` decoding by prefixing a :py:class:`PurePath` string with a special marker."""
        if isinstance(v, PurePath):
            if isinstance(v, Path):
                v = v.resolve()
            v = f'{SPECIAL_PATH_PREFIX}{v}'
        return super().dump_value(v=v)


class TomlFile:
    """Object that manages the getting and setting of TOML configurations.

    Houses an EventBus that allows you to subscribe Callables to changes in configuration.
    """

    def __init__(self, path: Path | str, default: dict[str, TOML_VALUE] | None = None) -> None:
        self.path = path
        # FIXME: Default not working as expected during import
        self._data: dict[str, TOML_VALUE | CommentValue] = default if default is not None else {}
        self._event_subscribers: dict[str, list[tuple[
            Callable[[...], None],  # Callable to call
            tuple[...],  # positional arguments
            dict[str, Any]  # keyword arguments
        ]]] = {}
        if self.reload() is False:
            warnings.warn(f'Could not load TOML file {self.path} on initialization.')

    def __getitem__(self, key: str) -> TOML_VALUE | None:
        return self.get_key(key)

    def __setitem__(self, key: str, value: TOML_VALUE) -> None:
        self.set_key(key, value)

    def __delitem__(self, key: str) -> None:
        del self._data[key]

    def _search_scope(self, path: str, mode: str) -> tuple[dict[str, TOML_VALUE | CommentValue], str]:
        """Search data for the given path to the value, and if found, return the scope and the key that path belongs to.

        :param path: Path to search data for.
        :param mode: Mode that determines which exceptions to raise.
        :return: Tuple containing the scope where the value is found, and the value key to access.
        :raises KeyError: If mode is 'set' and path is a table OR if mode is 'get' and path doesn't exist.
        :raises ValueError: If path is an empty string.
        """
        if not path:
            raise ValueError('Path can not be an empty string.')

        scope = self._data
        paths = path.split('/')

        if len(paths) > 1:
            for i, key in enumerate(paths):
                if key:
                    if i == len(paths) - 1:
                        if mode == 'set' and isinstance(scope.get(key), dict):
                            raise KeyError(f'Cannot reassign Table "{".".join(paths[:i])}" to variable.')
                        if mode == 'get' and key not in scope:
                            raise KeyError(f'Key "{key}" not in Table "{".".join(paths[:i]) or "/"}".')

                    elif isinstance(scope.get(key), dict):
                        scope = scope[key]
                        continue
        else:
            key = path

        # noinspection PyUnboundLocalVariable
        return scope, key

    @property
    def path(self) -> Path:
        """Current OS path that TomlFile with save and reload from."""
        return self._path

    @path.setter
    def path(self, value: Path | str) -> None:
        self._path = Path(value)

    def get_key(self, key: str) -> TOML_VALUE | None:
        """Get a key from path. Searches with each '/' defining a new table to check.

        Calls event $get:{key}.

        :param key: Key to get value from.
        """
        scope, path = self._search_scope(key, 'get')

        for sub in self._event_subscribers.get(f'$get:{key}', []):
            sub[0](*sub[1], **sub[2])

        # Get value from CommentValue
        val = scope.get(path)
        if isinstance(val, CommentValue):
            val = val.val

        return val

    def set_key(self, key: str, value: TOML_VALUE, comment: str | None = None) -> None:
        """Set a key at path. Searches with each '/' defining a new table to check.

        Calls event $set:{key}.

        :param key: Key to set.
        :param value: Value to set key as.
        :param comment: Append an optional comment to the value.
        """
        scope, path = self._search_scope(key, 'set')

        # Preserve comments, or edit them if comment argument was filled
        prev_val = scope.get(path)
        if isinstance(prev_val, CommentValue):
            prev_val.val = value
            value = prev_val
            if comment is not None:
                prev_val.comment = comment
        if not isinstance(value, CommentValue) and comment is not None:
            value = CommentValue(val=value, comment=f'# {comment}', beginline=True, _dict=dict)

        scope[path] = value

        for sub in self._event_subscribers.get(f'$set:{key}', []):
            sub[0](*sub[1], **sub[2])

    def save(self) -> bool:
        """Save current settings to self.path."""
        return self.export_to(self.path)

    def reload(self) -> bool:
        """Reset settings to settings stored in self.path."""
        for sub in self._event_subscribers.get('$reload', []):
            sub[0](*sub[1], **sub[2])
        return self.import_from(self.path, update=True)

    def export_to(self, path: Path | str) -> bool:
        """Export internal dictionary as a TOML file to path.

        :return: True if successful, otherwise False.
        """
        # TODO: upgrade similarly to import_from
        path = Path(path)  # Make sure path is of type Path
        if path.parent.is_dir():
            with path.open(mode='w', encoding='utf8') as file:
                toml.dump(self._data, file, encoder=BetterTomlEncoder())

            return True
        return False

    def import_from(self, path: Path | str, update: bool = False) -> bool:
        """Import TOML file from path to internal dictionary.

        :return: True if successful, otherwise False."""

        path = Path(path)  # Make sure path is of type Path
        if path.is_file():
            try:
                with path.open(mode='r', encoding='utf8') as file:
                    toml_data = toml.load(file, decoder=BetterTomlDecoder())
                    if update:
                        self._data.update(toml_data)
                    else:
                        self._data = toml_data

            except (LookupError, OSError, TomlDecodeError):
                pass  # Pass to end of function, to fail.

            else:
                # Execute all events subscribed to $set
                for name, event in self._event_subscribers.items():
                    if name.startswith('$set'):
                        for sub in event:
                            sub[0](*sub[1], **sub[2])

                return True

        # If failed:
        for sub in self._event_subscribers.get('$fail:import', []):
            sub[0](*sub[1], **sub[2])
        return False

    def hook_event(self, key: str, f: Callable, *args, **kwargs) -> None:
        """Hook a Callable to an event. The callable will be called with args and kwargs when the event is activated."""
        subscribers = self._event_subscribers.get(key, [])
        subscribers.append((f, args, kwargs))
        self._event_subscribers[key] = subscribers



###################################################################################################
#                              MIT Licence (C) 2022 Cubicpath@Github                              #
###################################################################################################
"""Utils for hi_getter."""
import json
import os
from collections.abc import Iterable
from collections.abc import Mapping
from http import HTTPStatus
from pathlib import Path

__all__ = (
    'dump_data',
    'HTTP_CODE_MAP',
    'unique_values',
)

# pylint: disable=not-an-iterable
HTTP_CODE_MAP = {status.value: (status.phrase, status.description) for status in HTTPStatus}


def unique_values(data: Iterable) -> set:
    """Recursively get all values in any Iterables. For Mappings, ignore keys and only remember values."""
    new: set = set()
    if isinstance(data, Mapping):
        # Loop through Mapping values
        for value in data.values():
            new.update(unique_values(value))
    elif isinstance(data, Iterable) and not isinstance(data, str):
        # Loop through Iterable values
        for value in data:
            new.update(unique_values(value))
    else:
        # Finally, get value
        new.add(data)
    return new


def dump_data(path: Path | str, data: bytes | dict | str, encoding: str | None = None) -> None:
    """Dump data to path as a file."""
    default_encoding = 'utf8'
    path = Path(path)
    if not path.parent.exists():
        os.makedirs(path.parent)

    if isinstance(data, str):
        path.write_text(data, encoding=encoding or default_encoding)
    elif isinstance(data, bytes):
        if encoding is not None:
            data = data.decode(encoding=encoding)
            path.write_text(data, encoding=encoding)
        else:
            path.write_bytes(data)
    elif isinstance(data, dict):
        with path.open('w', encoding=encoding or default_encoding) as file:
            json.dump(data, file, indent=2)



###################################################################################################
#                              MIT Licence (C) 2022 Cubicpath@Github                              #
###################################################################################################
"""Module containing code related to translation and language files."""
import json
from pathlib import Path
from string import ascii_letters
from string import digits
from typing import Annotated

from .constants import RESOURCE_PATH

LANG_PATH: Path = RESOURCE_PATH / 'lang'
"""Directory containing language JSON data."""


class Language:
    """Object containing language data. Retrieves stored data with associated RFC 5646 language tags."""

    def __init__(self, primary: Annotated[str, 2] | None = None, region: Annotated[str, 2] | None = None, **kwargs) -> None:
        """Create a new Language object with data from the associated language tag built using the given arguments.
        For more information on RFC 5646, visit https://datatracker.ietf.org/doc/html/rfc5646

        :param primary: The first part of the language code. i.e. [en]-US | Section 2.2.1
        :param region: The (commonly) second part of the language code. i.e. en-[US] | Section 2.2.4

        :keyword ext_lang: Extension to primary subtag | Section 2.2.2
        :keyword script: Script used to write the language | Section 2.2.3
        :keyword variants: A variant of the given language subtag | Section 2.2.5
        :keyword extensions: Extension to the given language, prefixed with a singleton | Section 2.2.6
        :keyword private_use: Any subtag that is not publicly defined | Section 2.2.7
        :type ext_lang: Annotated[str, 3] | None
        :type script: Annotated[str, 4] | None = None
        :type variants: Sequence[str] | None
        :type extensions: Sequence[tuple[Annotated[str, 1], str]] | None
        :type private_use: Sequence[str] | None

        :raises ValueError: If a given subtag is invalid.
        """

        (ext_lang, script, variants, extensions, private_use) = (
            kwargs.pop('ext_lang', None),
            kwargs.pop('script', None),
            kwargs.pop('variants', None),
            kwargs.pop('extensions', None),
            kwargs.pop('private_use', None),
        )

        self._data: dict[str, str] = {}
        self.tag:  str = ''

        sub_tags:  list[str] = []

        if primary is None and ext_lang is None and private_use is None:
            raise ValueError('The primary and/or ext_lang and/or private_use subtag must be filled out.')

        if primary is not None:
            primary = primary.lower()

            # RFC 5646 section 2.2.1.1 and 2.2.1.2
            if not primary.isalpha() or not 3 >= len(primary) >= 2:
                raise ValueError(f'Primary language subtag "{primary}" is not valid.')

            sub_tags.append(primary)

        if ext_lang is not None:
            ext_lang = ext_lang.lower()

            # RFC 5646 section 2.2.2.1
            if not ext_lang.isalpha() or not len(ext_lang) == 3:
                raise ValueError(f'Extended language subtag "{ext_lang}" is not valid.')

            sub_tags.append(ext_lang)

        if script is not None:
            script = script.title()

            # RFC 5646 section 2.2.3.2
            if not script.isalpha() or not len(script) == 4:
                raise ValueError(f'Script subtag "{script}" is not valid.')

            sub_tags.append(script)

        if region is not None:
            region = region.upper()

            # ISO 3166-1 or UN M.49
            if not (len(region) == 2 and region.isalpha()) and not (len(region) == 5 and region[:2].isalpha() and region[2:].isnumeric()):
                raise ValueError(f'Region subtag "{region}" is not valid.')

            sub_tags.append(region)

        if variants is not None:
            checked = []
            for variant in variants:

                # RFC 5646 section 2.2.5.4
                if not (variant[0] in ascii_letters and 8 >= len(variant) >= 5) or not (variant[0] in digits and 8 >= len(variant) >= 4):
                    raise ValueError(f'Variant subtag "{variant}" is not valid.')

                # RFC 5646 section 2.2.5.5
                if variant in checked:
                    raise ValueError(f'Variant subtag "{variant}" is repeated.')

                checked.append(variant)
                sub_tags.append(variant)

        if extensions is not None:
            checked = []
            for singleton, extension in extensions:
                extension = extension.lower()

                if not (len(singleton) == 1 and (singleton in digits or singleton in ascii_letters)):
                    raise ValueError(f'Singleton subtag "{singleton}" is not valid.')

                # RFC 5646 section 2.2.6.3
                if singleton in checked:
                    raise ValueError(f'Singleton subtag "{singleton}" is repeated.')

                # RFC 5646 section 2.2.6.5
                if not (extension.isalnum() and 8 >= len(extension) >= 2):
                    raise ValueError(f'Extension subtag "{extension}" is not valid.')

                checked.append(singleton)
                sub_tags.extend((singleton, extension))

        if private_use is not None:
            sub_tags.append('x')
            for i, private_sub_tag in enumerate(private_use):
                private_sub_tag = private_sub_tag.lower()

                # If previous subtag was not a singleton
                if i != 0 and len(private_use[i - 1]) != 1:
                    if len(private_sub_tag) == 4:
                        private_sub_tag = private_sub_tag.title()
                    elif len(private_sub_tag) == 2:
                        private_sub_tag = private_sub_tag.upper()

                if not (private_sub_tag.isalnum() and 8 >= len(private_sub_tag) >= 1):
                    raise ValueError(f'The private subtag "{private_sub_tag}" is not valid.')

                sub_tags.append(private_sub_tag)

        self.tag = '-'.join(sub_tags)

        for lang_file in LANG_PATH.iterdir():
            if lang_file.suffix == '.json' and lang_file.with_suffix('').name.lower() == self.tag.replace('-', '_').lower():
                self._data.update(json.loads(lang_file.read_text(encoding='utf8')))
                break

    def __repr__(self) -> str:
        return f'<Language data for {self.tag}>'

    def __getitem__(self, key) -> str:
        return self.get_raw(key)

    def get(self, key: str, *args, default: str | None = None) -> str | None:
        """Get a translation key and format with the given arguments.

        :param key: Key to get in JSON data.
        :param args: Formatting arguments; packed into tuple and used with str modulus.
        :param default: Default string if __key doesn't exist. (Is not formatted)
        """
        result = self._data.get(key, default)
        if result is not default:
            result = result % args

        return result

    def get_raw(self, key: str) -> str:
        """Get raw result, no formatting or defaults.

        :raises KeyError: If key is not in JSON data.
        """
        return self._data[key]
